"""
Manager for virtual environments
================================
"""

import datetime
import functools
import os
import shutil
import subprocess
import tempfile

import venv
import click

from . import config


def locate(venv_name, bin_):
    """
    @brief Path to venv binaries
    """
    return os.path.realpath(
        os.path.join(
            config.VENV_DIR,
            venv_name,
            "bin",
            bin_))


def isvenv(venv_name):
    """
    @brief Whether a venv with such a name exists
    """
    for f in ["activate", "python", "pip"]:
        if not os.path.isfile(locate(venv_name, f)):
            return False
    return True


def default_venv_name(file_name):
    """
    @brief Default name for a venv
    """
    abs_ = os.path.abspath(file_name)
    return abs_.replace(os.path.sep, "_").replace(".", "_").strip("_")


def create(venv_name, reqs=None):
    """
    @brief Create a new virtual environment
    """
    if isvenv(venv_name) and not click.confirm(
            f"'{venv_name}' already exists. Continue?"):
        return

    env = venv.EnvBuilder(with_pip=True)
    dir_env = os.path.join(config.VENV_DIR, venv_name)
    env.create(dir_env)
    config.set(config.TIME, venv_name, datetime.datetime.now())

    if reqs is None and os.path.isfile("requirements.txt"):
        reqs = "requirements.txt"

    if reqs is not None:
        pip(venv_name, "install", "-r", reqs)


def ls(venv_names, long_=False):
    """
    @brief List known virtual environments
    """
    if not venv_names:
        venv_names = os.listdir(config.VENV_DIR)

    for venv_name in sorted(venv_names):

        yield click.style(venv_name, fg='green', bold=True)

        if time := config.get(config.TIME).get(venv_name):
            yield click.style(f"created at {time}", fg='green', italic=True)

        if annotation := config.get(config.ANNOTATION).get(venv_name):
            for line in annotation.split("\n"):
                yield click.style(line, fg='green', italic=True)

        if not isvenv(venv_name):
            yield click.style("broken", fg='red', bold=True)

        a = click.format_filename(locate(venv_name, "activate"))
        yield click.style("activate: " + click.style(f"source {a}", fg='red'))

        file_names = [
            click.format_filename(k) for k,
            v in config.get(
                config.ASSOCIATION).items() if v == venv_name]

        if file_names:
            yield click.style("used by: " + click.style(", ".join(file_names), fg='red'))
        else:
            yield click.style("no known users", fg='red')

        if long_:
            try:
                res = pip(
                    venv_name,
                    "list",
                    verbose=False,
                    stdout=subprocess.PIPE)
                table = res.stdout.decode('utf-8')
                for line in table.split("\n"):
                    yield click.style(line)
            except RuntimeError as r:
                yield click.style(r, fg='red', bold=True)

        yield click.style("")


def create_if_not_exist(venv_name):
    """
    @brief Create a venv if it does not exist
    """
    if not isvenv(venv_name):
        if click.confirm(f"Create new venv '{venv_name}'?"):
            create(venv_name)
        else:
            raise RuntimeError(f"Does not exist - {venv_name}")


def venv_must_exist(func):
    """
    @brief Raise error if venv does not exist
    """
    @functools.wraps(func)
    def with_venv_must_exist(venv_name, *args, **kwargs):
        create_if_not_exist(venv_name)
        return func(venv_name, *args, **kwargs)
    return with_venv_must_exist


@venv_must_exist
def associate(venv_name, file_names):
    """
    @brief Associate a venv with file names
    """
    for file_name in file_names:
        config.set(config.ASSOCIATION, os.path.abspath(file_name), venv_name)


@venv_must_exist
def annotate(venv_name):
    """
    @brief Leave an annotation about a venv
    """
    marker = f'# Annotation for {venv_name}\n'
    start = config.get(config.ANNOTATION).get(venv_name, marker)
    message = click.edit(start)

    if message is not None:
        annotation = message.split(marker, 1)[-1].strip()
        config.set(config.ANNOTATION, venv_name, annotation)


def add(venv_name):
    """
    @brief Add a venv created outside of vm
    """
    os.symlink(
        os.path.abspath(venv_name),
        os.path.join(
            config.VENV_DIR,
            os.path.basename(venv_name)))
    config.set(config.TIME, venv_name, datetime.datetime.now())


def rm(venv_name):
    """
    @brief Removes a venv
    """
    d = f"{config.VENV_DIR}/{venv_name}"
    if click.confirm(f"Are you sure you wish to remove '{d}'?"):
        shutil.rmtree(d)


def python(venv_name, file_name, *args, verbose=True, **kwargs):
    """
    @brief Run a Python file in a virtual environment
    """
    abs_file_name = os.path.abspath(file_name)

    if file_name is None and venv_name is None:
        raise RuntimeError("Require venv name or file name")

    if venv_name is None:
        venv_name = config.get(
            config.ASSOCIATION).get(
            abs_file_name,
            default_venv_name(file_name))

    create_if_not_exist(venv_name)

    config.set(config.ASSOCIATION, abs_file_name, venv_name)

    bin_python = locate(venv_name, "python")

    if verbose:
        click.secho(
            f"python = {click.format_filename(bin_python)}",
            fg='green',
            bold=True)
    return subprocess.run([bin_python] + list(args) + [file_name], check=False, **kwargs)


@venv_must_exist
def pip(venv_name, *cmds, verbose=True, **kwargs):
    """
    @brief Run a Python file in a virtual environment
    """
    bin_pip = locate(venv_name, "pip")

    if verbose:
        click.secho(
            f"pip = {click.format_filename(bin_pip)}",
            fg='green',
            bold=True)

    return subprocess.run([bin_pip] + list(cmds), check=False, **kwargs)


@venv_must_exist
def activate(venv_name):
    """
    Activate venv in new shell
    """
    activation = locate(venv_name, "activate")

    bashrc = f"""
    #!/bin/bash

    source $HOME/.bashrc
    source {activation}
    """

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as tf:
        tf.write(bashrc)
        command = f'bash --rcfile {tf.name}'

    os.system(command)



"""
Config file for venvs
=====================
"""

import os
import configparser
from contextlib import contextmanager


VENV_DIR = os.path.join(os.environ['HOME'], '.venvmgr')
CONFIG_FILE = os.path.join(os.environ['HOME'], '.venvmgr.conf')
ASSOCIATION = "ASSOCIATION"
ANNOTATION = "ANNOTATION"
TIME = "TIME"


@contextmanager
def open_config():
    """
    @brief Context to open and save config file
    """
    os.makedirs(VENV_DIR, exist_ok=True)
    with open(CONFIG_FILE, 'a'):
        pass

    config = configparser.ConfigParser()
    config.read(CONFIG_FILE)

    yield config

    with open(CONFIG_FILE, 'w') as f:
        config.write(f)


def set(section, key, value):
    """
    @brief Alter the config
    """
    with open_config() as o:
        if not o.has_section(section):
            o.add_section(section)
        o[section][key] = str(value)


def get(section):
    """
    @brief Alter the config
    """
    with open_config() as o:

        if not o.has_section(section):
            return {}

        return o[section]



"""
CLI for venv manager
====================
"""


import os
import click

from . import venvmgr
from . import config


__version__ = "0.1.1"


def get_venv_names(ctx, args, incomplete):
    return os.listdir(config.VENV_DIR)


@click.group()
@click.version_option(__version__)
def commands():
    pass


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('file', required=False, default=None,
                type=click.Path(exists=True))
@click.option('--venv',
              type=click.STRING,
              default=None,
              required=False,
              shell_complete=get_venv_names,
              help="Name of venv")
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def python(file, venv, args):
    """
    Python though with a venv argument

    If no VENV provided but a FILE is provided, uses a VENV inferred from FILE
    """
    venvmgr.python(venv, file, *args)


@click.command()
@click.option('-l', '--long', type=bool, default=False,
              is_flag=True, help='long listing format')
@click.argument('name', nargs=-1, required=False,
                type=click.STRING, shell_complete=get_venv_names)
def ls(long, name):
    """
    List information about venvmgr known venvs
    """
    click.echo_via_pager("\n".join(venvmgr.ls(name, long)))


@click.command()
@click.argument('name', type=click.STRING, shell_complete=get_venv_names)
@click.argument('file', type=click.Path(exists=True), nargs=-1)
def associate(name, file):
    """
    Associate venv NAME with FILE
    """
    venvmgr.associate(name, file)


@click.command()
@click.argument('name', type=click.STRING, shell_complete=get_venv_names)
@click.option('-r', '--requirement',
              type=click.Path(exists=True),
              default=None,
              required=False,
              help="Install from the given requirements file")
def create(name, requirement):
    """
    Create NAME venv
    """
    venvmgr.create(name, requirement)


@click.command()
@click.argument('name', type=click.STRING, shell_complete=get_venv_names)
def rm(name):
    """
    Remove NAME venv
    """
    venvmgr.rm(name)


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.option('--venv', type=click.STRING, required=True,
              shell_complete=get_venv_names, help="Name of venv")
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def pip(venv, args):
    """
    Pip though with venv argument
    """
    venvmgr.pip(venv, *args)


@click.command()
@click.argument('dir', type=click.STRING)
def add(dir):
    """
    Add a venv from DIR created outside of vm
    """
    venvmgr.add(dir)


@click.command()
@click.argument('name', type=click.STRING, shell_complete=get_venv_names)
def annotate(name):
    """
    Annotate NAME venv
    """
    venvmgr.annotate(name)


@click.command()
def home():
    """
    Location of venvmgr home directory
    """
    click.echo(config.VENV_DIR)


@click.command()
def dotfile():
    """
    Location of venvmgr configuration file
    """
    click.echo(config.CONFIG_FILE)


@click.command()
@click.argument('name', type=click.STRING, shell_complete=get_venv_names)
def activate(name):
    """
    Activate NAME venv
    """
    venvmgr.activate(name)


commands.add_command(create)
commands.add_command(rm)
commands.add_command(python)
commands.add_command(ls)
commands.add_command(pip)
commands.add_command(add)
commands.add_command(annotate)
commands.add_command(associate)
commands.add_command(home)
commands.add_command(dotfile)
commands.add_command(activate)
